/*
    Filter->de- for VirtualDub (0.2).
		Based on Red/Green/Blue (RGB) Adjustment Filter for VirtualDub.(1.0) by Donald A. Graft
    Copyright (C) 1999-2000 Donald A. Graft
	Copyright (C) 2004 Eugene Khoroshavin

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

	The author can be contacted at:
	Eugene Khoroshavin
	fdump@mail.ru
*/

#include <windows.h>
#include <commctrl.h>
#include <stdio.h>
#include <crtdbg.h>
#include <math.h>
#include <stdlib.h>

#include "ScriptInterpreter.h"
#include "ScriptError.h"
#include "ScriptValue.h"

#include "resource.h"
#include "filter.h"

///////////////////////////////////////////////////////////////////////////

int RunProc(const FilterActivation *fa, const FilterFunctions *ff);
int StartProc(FilterActivation *fa, const FilterFunctions *ff);
int EndProc(FilterActivation *fa, const FilterFunctions *ff);
long ParamProc(FilterActivation *fa, const FilterFunctions *ff);
int InitProc(FilterActivation *fa, const FilterFunctions *ff);
int ConfigProc(FilterActivation *fa, const FilterFunctions *ff, HWND hwnd);
void StringProc(const FilterActivation *fa, const FilterFunctions *ff, char *str);
void ScriptConfig(IScriptInterpreter *isi, void *lpVoid, CScriptValue *argv, int argc);
bool FssProc(FilterActivation *fa, const FilterFunctions *ff, char *buf, int buflen);
int normpos( int a, int b, int c, int pos, long width);
void normrgb(long *r, long *g, long *b);
void normc(long *c);
int normv(int v, int norm);
UCHAR valtoch(int v, int ch);
int chtoval(UCHAR ch1, UCHAR ch2);
///////////////////////////////////////////////////////////////////////////
#define u22 (16384)
#define TTH (10000)

const char * f11txt = "Red channel Shift";
const char * f12txt = "Green channel Shift";
const char * f13txt = "Blue channel Shift";
const char * f1ntxt = "Chroma Shift";
const char * f1ef1txt = "Pt chain Shift";
const char * f1ef2txt = "hor. Shift";
const char * f1ef3txt = "ver. Shift";
const char * f2txt = "De-border";
const char * f2ef1txt = "Smoothness";
const char * f2ef2txt = "Roughness";
const char * f2ef3txt = "Threshold";
const char * f3txt = "De-blocking";
const char * f3ef1txt = "Threshold";
const char * f3ef2txt = "Base X-Shift";
const char * f3ef3txt = "Base X-Shift";




typedef struct MyFilterData {
	IFilterPreview		*ifp;

	int					rshpt;
	int					rshxx;
	int					rshyy;

	int					gshpt;
	int					gshxx;
	int					gshyy;

	int					bshpt;
	int					bshxx;
	int					bshyy;

	int					rshxy;
	int					gshxy;
	int					bshxy;

	int					dbldif;
	int					dbhdif;
	int					dbhdam;

	int					dbctrh;
	int					dbcxsh;
	int					dbcysh;
	int					dbcbas;

	int					shiFlg;
	int					keeppt;
	int					showpt;
	int					cascad;

	int					rshlock;
	int					gshlock;
	int					bshlock;
	int					dbllock;
	int					dbclock;
	int					shiunion;
	int					nmfunc;

	long				salign;
	long				dalign;
	long				width;
	long				height;
} MyFilterData;

////////////////////////////////////////////////////////////////////////////////
void setFlg (MyFilterData *mfd);
void shiftch(MyFilterData *mfd, long *r, long *g, long *b, Pixel32 *src,  int x, int y);
///////////////////////////////////////////////////////////////////////////////

bool FssProc(FilterActivation *fa, const FilterFunctions *ff, char *buf, int buflen) {
	MyFilterData *mfd = (MyFilterData *)fa->filter_data;
	char mary[37];
	mary[0]		= 34;
	mary[1]		= valtoch(mfd->rshpt, 1);
	mary[2]		= valtoch(mfd->rshpt, 2);
	mary[3]		= valtoch(mfd->rshxx, 1);
	mary[4]		= valtoch(mfd->rshxx, 2);
	mary[5]		= valtoch(mfd->rshyy, 1);
	mary[6]		= valtoch(mfd->rshyy, 2);

	mary[7]		= valtoch(mfd->gshpt, 1);
	mary[8]		= valtoch(mfd->gshpt, 2);
	mary[9]		= valtoch(mfd->gshxx, 1);
	mary[10]	= valtoch(mfd->gshxx, 2);
	mary[11]	= valtoch(mfd->gshyy, 1);
	mary[12]	= valtoch(mfd->gshyy, 2);

	mary[13]	= valtoch(mfd->bshpt, 1);
	mary[14]	= valtoch(mfd->bshpt, 2);
	mary[15]	= valtoch(mfd->bshxx, 1);
	mary[16]	= valtoch(mfd->bshxx, 2);
	mary[17]	= valtoch(mfd->bshyy, 1);
	mary[18]	= valtoch(mfd->bshyy, 2);

	mary[19]	= valtoch(mfd->dbldif, 1);
	mary[20]	= valtoch(mfd->dbldif, 2);
	mary[21]	= valtoch(mfd->dbhdif, 1);
	mary[22]	= valtoch(mfd->dbhdif, 2);
	mary[23]	= valtoch(mfd->dbhdam, 1);
	mary[24]	= valtoch(mfd->dbhdam, 2);

	mary[25]	= valtoch(mfd->dbctrh, 1);
	mary[26]	= valtoch(mfd->dbctrh, 2);
	mary[27]	= valtoch(mfd->dbcxsh, 1);
	mary[28]	= valtoch(mfd->dbcxsh, 2);
	mary[29]	= valtoch(mfd->dbcysh, 1);
	mary[30]	= valtoch(mfd->dbcysh, 2);
	mary[31]	= valtoch(mfd->dbcbas, 1);
	mary[32]	= valtoch(mfd->dbcbas, 2);

	mary[33]	= 0x80			| // not use 8th bit
		(mfd->keeppt	    )	|
		(mfd->rshlock	<< 1)	|
		(mfd->gshlock	<< 2)	|
		(mfd->bshlock	<< 3)	|
		(mfd->shiunion	<< 4)	|
		(mfd->showpt	<< 5)	|
		(mfd->cascad	<< 6)	;
	mary[34] = 0x80				| // not use 8th bit
		(mfd->dbllock	    )	|
		(mfd->dbclock	<< 1)	;
	mary[35] = 34;
	mary[36] = 0;
	_snprintf(buf, buflen, "Config(%s)", mary, 37);

	return true;
}

void ScriptConfig(IScriptInterpreter *isi, void *lpVoid, CScriptValue *argv, int argc) {
	FilterActivation *fa = (FilterActivation *)lpVoid;
	MyFilterData *mfd = (MyFilterData *)fa->filter_data;

	char **mar = argv[0].asString();
	UCHAR *mary = (UCHAR *)*mar;
	mfd->rshpt			= chtoval(mary[0], mary[1]);
	mfd->rshxx			= chtoval(mary[2], mary[3]);
	mfd->rshyy			= chtoval(mary[4], mary[5]);

	mfd->gshpt			= chtoval(mary[6], mary[7]);
	mfd->gshxx			= chtoval(mary[8], mary[9]);
	mfd->gshyy			= chtoval(mary[10], mary[11]);

	mfd->bshpt			= chtoval(mary[12], mary[13]);
	mfd->bshxx			= chtoval(mary[14], mary[15]);
	mfd->bshyy			= chtoval(mary[16], mary[17]);

	mfd->dbldif			= chtoval(mary[18], mary[19]);
	mfd->dbhdif			= chtoval(mary[20], mary[21]);
	mfd->dbhdam			= chtoval(mary[22], mary[23]);

	mfd->dbctrh			= chtoval(mary[24], mary[25]);
	mfd->dbcxsh			= chtoval(mary[26], mary[27]);
	mfd->dbcysh			= chtoval(mary[28], mary[29]);
	mfd->dbcbas			= chtoval(mary[30], mary[31]);

	int flgs = mary[32];
	mfd->keeppt			= ((flgs & 0x01) > 0);
	mfd->rshlock		= ((flgs & 0x02) > 0);
	mfd->gshlock		= ((flgs & 0x04) > 0);
	mfd->bshlock		= ((flgs & 0x08) > 0);
	mfd->shiunion		= ((flgs & 0x10) > 0);
	mfd->showpt			= ((flgs & 0x20) > 0);
	mfd->cascad			= ((flgs & 0x40) > 0);
	flgs = mary[33];
	mfd->dbllock		= ((flgs & 0x01) > 0);
	mfd->dbclock		= ((flgs & 0x02) > 0);

	mfd->nmfunc = IDC_RSH;
	setFlg(mfd);
	mfd->nmfunc = IDC_DBL;
	setFlg(mfd);
	mfd->nmfunc = IDC_DBR;
	setFlg(mfd);
	mfd->nmfunc = IDC_RSH;
	setFlg(mfd);
}

ScriptFunctionDef func_defs[]={
	{ (ScriptFunctionPtr)ScriptConfig, "Config", "0s" },
	{ NULL },
};

CScriptObject script_obj={
	NULL, func_defs
};

struct FilterDefinition filterDef_tutorial = {

	NULL, NULL, NULL,		// next, prev, module
	"filter->de- (v0.2)",	// name
	"Support shift chroma or R,G,B channels, and de-border",
							// desc
	"Eugene Khoroshavin", 		// maker
	NULL,					// private_data
	sizeof(MyFilterData),	// inst_data_size

	InitProc,				// initProc
	NULL,					// deinitProc
	RunProc,				// runProc
	NULL,					// paramProc
	ConfigProc, 			// configProc
	StringProc, 			// stringProc
	StartProc,				// startProc
	EndProc,				// endProc

	&script_obj,			// script_obj
	FssProc,				// fssProc

};

int StartProc(FilterActivation *fa, const FilterFunctions *ff)
{
	MyFilterData *mfd = (MyFilterData *)fa->filter_data;
	return 0;
}

int EndProc(FilterActivation *fa, const FilterFunctions *ff)
{
	MyFilterData *mfd = (MyFilterData *)fa->filter_data;
	return 0;
}

///////////////////////////////////////////////////////////////////////////

int RunProc(const FilterActivation *fa, const FilterFunctions *ff)
{
	MyFilterData *mfd = (MyFilterData *)fa->filter_data;
	const PixDim	width = fa->src.w;
	mfd->width = width;
	const PixDim	height = fa->src.h;
	mfd->height = height;
	Pixel32 *src, *dst, *nwpic, *nwpics;
	int x, y;
	long rs, gs, bs, r[16], g[16], b[16], dch[4], dcv[4];
	UCHAR *srcl;
	mfd->salign = (fa->src.pitch - width * 4);
	mfd->dalign = (fa->dst.pitch - width * 4);
	const PixDim vol = fa->src.w * (fa->src.h + 1);
	nwpics = new Pixel32 [vol];

	src = fa->src.data;
	dst = fa->dst.data;
	nwpic = nwpics;
	if (mfd->shiFlg){
		for (y = 0; y < height; y++)
		{
			for (x = 0; x < width; x++)
			{
				bs = ((UCHAR *)src)[0];
				gs = ((UCHAR *)src)[1];
				rs = ((UCHAR *)src)[2];
				shiftch(mfd, &rs, &gs, &bs, src, x, y);
				dst[0] = ((int) rs << 16) | ((int) gs << 8) | ((int) bs);
				nwpic[0] = dst[0];
				src++;
				dst++;
				nwpic++;
			}
			src	= (Pixel32 *)((char *)src + mfd->salign);
			dst	= (Pixel32 *)((char *)dst + mfd->dalign);
		}
	}
	else{
		for (y = 0; y < height; y++)
		{
			for (x = 0; x < width; x++)
			{
				dst[0] = src[0];
				nwpic[0] = dst[0];
				src++;
				dst++;
				nwpic++;
			}
			src	= (Pixel32 *)((char *)src + mfd->salign);
			dst	= (Pixel32 *)((char *)dst + mfd->dalign);
		}
	}
/*
c d | e f
8 9 | a b
----|----
4 5 | 6 7
0 1 | 2 3
*/
	long pwid;
	Pixel32 *nnn;
	if (!mfd->dbldif) goto nodbl;
	dst = fa->dst.data;
	dst	= (Pixel32 *)((char *)dst + fa->dst.pitch + 4);
	nwpic = nwpics + mfd->width + 1;
	for (y = 1; y < (height - 2); y++)
	{
		for (x = 1; x < (width - 2); x++)
		{
			if (mfd->showpt && !mfd->cascad) {
				nnn = nwpic;
				pwid = mfd->width * 4;
			}
			else {
				nnn = dst;
				pwid = fa->dst.pitch;
			}
				srcl = (UCHAR *)nnn - pwid - 4; 
				b[0] = srcl[0];
				g[0] = srcl[1];
				r[0] = srcl[2];
				srcl = (UCHAR *)nnn - pwid;
				b[1] = srcl[0];
				g[1] = srcl[1];
				r[1] = srcl[2];
				srcl = (UCHAR *)nnn - pwid + 4;
				b[2] = srcl[0];
				g[2] = srcl[1];
				r[2] = srcl[2];
				srcl = (UCHAR *)nnn - pwid + 8;
				b[3] = srcl[0];
				g[3] = srcl[1];
				r[3] = srcl[2];
				srcl = (UCHAR *)nnn - 4;
				b[4] = srcl[0];
				g[4] = srcl[1];
				r[4] = srcl[2];
				srcl = (UCHAR *)nnn;
				b[5] = srcl[0];
				g[5] = srcl[1];
				r[5] = srcl[2];
				srcl = (UCHAR *)nnn + 4;
				b[6] = srcl[0];
				g[6] = srcl[1];
				r[6] = srcl[2];
				srcl = (UCHAR *)nnn + 8;
				b[7] = srcl[0];
				g[7] = srcl[1];
				r[7] = srcl[2];
				srcl = (UCHAR *)nnn + pwid - 4;
				b[8] = srcl[0];
				g[8] = srcl[1];
				r[8] = srcl[2];
				srcl = (UCHAR *)nnn + pwid;
				b[9] = srcl[0];
				g[9] = srcl[1];
				r[9] = srcl[2];
				srcl = (UCHAR *)nnn + pwid + 4;
				b[10] = srcl[0];
				g[10] = srcl[1];
				r[10] = srcl[2];
				srcl = (UCHAR *)nnn + pwid + 8;
				b[11] = srcl[0];
				g[11] = srcl[1];
				r[11] = srcl[2];
				srcl = (UCHAR *)nnn + pwid * 2 - 4;
				b[12] = srcl[0];
				g[12] = srcl[1];
				r[12] = srcl[2];
				srcl = (UCHAR *)nnn + pwid * 2;
				b[13] = srcl[0];
				g[13] = srcl[1];
				r[13] = srcl[2];
				srcl = (UCHAR *)nnn + pwid * 2 + 4;
				b[14] = srcl[0];
				g[14] = srcl[1];
				r[14] = srcl[2];
				srcl = (UCHAR *)nnn + pwid * 2 + 8;
				b[15] = srcl[0];
				g[15] = srcl[1];
				r[15] = srcl[2];

				dcv[0] = abs(r[1] - r[5]) +  abs(g[1] - g[5]) + abs(b[1] - b[5]);
				dcv[1] = abs(r[2] - r[6]) +  abs(g[2] - g[6]) + abs(b[2] - b[6]);
				dcv[2] = abs(r[9] - r[13]) +  abs(g[9] - g[13]) + abs(b[9] - b[13]);
				dcv[3] = abs(r[10] - r[14]) +  abs(g[10] - g[14]) + abs(b[10] - b[14]);
				dch[0] = abs(r[1] - r[2]) +  abs(g[1] - g[2]) + abs(b[1] - b[2]);
				dch[1] = abs(r[5] - r[6]) +  abs(g[5] - g[6]) + abs(b[5] - b[6]);
				dch[2] = abs(r[9] - r[10]) +  abs(g[9] - g[10]) + abs(b[9] - b[10]);
				dch[3] = abs(r[13] - r[14]) +  abs(g[13] - g[14]) + abs(b[13] - b[14]);
				
				long contr = 100 + mfd->dbhdam;
				bool borderv = 
					  (dcv[0] < mfd->dbldif)
					&& (dcv[1] < mfd->dbldif)
					&& (dcv[2] < mfd->dbldif)
					&& (dcv[3] < mfd->dbldif)
					&& (dch[0] > mfd->dbhdif)
					&& (dch[1] > mfd->dbhdif)
					&& (dch[2] > mfd->dbhdif)
					&& (dch[3] > mfd->dbhdif)
					&& (dch[0] < contr)
					&& (dch[1] < contr)
					&& (dch[2] < contr)
					&& (dch[3] < contr);

				dcv[0] = abs(r[4] - r[5]) +  abs(g[4] - g[5]) + abs(b[4] - b[5]);
				dcv[1] = abs(r[6] - r[7]) +  abs(g[6] - g[7]) + abs(b[6] - b[7]);
				dcv[2] = abs(r[8] - r[9]) +  abs(g[8] - g[9]) + abs(b[8] - b[9]);
				dcv[3] = abs(r[10] - r[11]) +  abs(g[10] - g[11]) + abs(b[10] - b[11]);
				dch[0] = abs(r[4] - r[8]) +  abs(g[4] - g[8]) + abs(b[4] - b[8]);
				dch[1] = abs(r[5] - r[9]) +  abs(g[5] - g[9]) + abs(b[5] - b[9]);
				dch[2] = abs(r[6] - r[10]) +  abs(g[6] - g[10]) + abs(b[6] - b[10]);
				dch[3] = abs(r[7] - r[11]) +  abs(g[7] - g[11]) + abs(b[7] - b[11]);

				bool borderh = 
					  ((dcv[0] < mfd->dbldif)
					&& (dcv[1] < mfd->dbldif)
					&& (dcv[2] < mfd->dbldif)
					&& (dcv[3] < mfd->dbldif)
					&& (dch[0] > mfd->dbhdif)
					&& (dch[1] > mfd->dbhdif)
					&& (dch[2] > mfd->dbhdif)
					&& (dch[3] > mfd->dbhdif)
					&& (dch[0] < contr)
					&& (dch[1] < contr)
					&& (dch[2] < contr)
					&& (dch[3] < contr));

				if (borderv || borderh) {
					if (mfd->showpt && !mfd->cascad) {
						Pixel32 *lsrc;
						if (borderv && borderh) {
							lsrc = (Pixel32*)((UCHAR*)dst - fa->dst.pitch);
							lsrc[0] = 0xffff; 
							dst[0] = 0xffff;
							lsrc = (Pixel32*)((UCHAR*)dst + fa->dst.pitch);
							lsrc[0] = 0xffff; 
							lsrc = (Pixel32*)((UCHAR*)dst + fa->dst.pitch * 2);
							lsrc[0] = 0xffff;
							lsrc = (Pixel32*)((UCHAR*)dst - 4);
							lsrc[0] = 0xffff; 
							lsrc = (Pixel32*)((UCHAR*)dst + 4);
							lsrc[0] = 0xffff; 
							lsrc = (Pixel32*)((UCHAR*)dst + 8);
							lsrc[0] = 0xffff;
						}
						else if (borderv) {
							lsrc = (Pixel32*)((UCHAR*)dst - fa->dst.pitch);
							lsrc[0] = 0xffff; 
							dst[0] = 0xffff;
							lsrc = (Pixel32*)((UCHAR*)dst + fa->dst.pitch);
							lsrc[0] = 0xffff; 
							lsrc = (Pixel32*)((UCHAR*)dst + fa->dst.pitch * 2);
							lsrc[0] = 0xffff;
						}
						else if (borderh) {
							dst[0] = 0xffff;
							lsrc = (Pixel32*)((UCHAR*)dst - 4);
							lsrc[0] = 0xffff; 
							lsrc = (Pixel32*)((UCHAR*)dst + 4);
							lsrc[0] = 0xffff; 
							lsrc = (Pixel32*)((UCHAR*)dst + 8);
							lsrc[0] = 0xffff;
						}
					}
					else {
						if (borderv && borderh) {
							srcl = (UCHAR *)dst - fa->dst.pitch;
							srcl[0] = (UCHAR)((3 * b[1] + b[2]) >> 2);
							srcl[1] = (UCHAR)((3 * g[1] + g[2]) >> 2);
							srcl[2] = (UCHAR)((3 * r[1] + r[2]) >> 2);

							srcl = (UCHAR *)dst - fa->dst.pitch + 4;
							srcl[0] = (UCHAR)((b[2] + 3 * b[1]) >> 2);
							srcl[1] = (UCHAR)((g[2] + 3 * g[1]) >> 2);
							srcl[2] = (UCHAR)((r[2] + 3 * r[1]) >> 2);

							srcl = (UCHAR *)dst - 4;
							srcl[0] = (UCHAR)((3 * b[4] + b[8]) >> 2);
							srcl[1] = (UCHAR)((3 * g[4] + g[8]) >> 2);
							srcl[2] = (UCHAR)((3 * r[4] + r[8]) >> 2);

							srcl = (UCHAR *)dst;
							srcl[0] = (UCHAR)((5 * b[5] + b[6] + b[9] + b[10]) >> 3);
							srcl[1] = (UCHAR)((5 * g[5] + g[6] + g[9] + g[10]) >> 3);
							srcl[2] = (UCHAR)((5 * r[5] + r[6] + r[9] + r[10]) >> 3);

							srcl = (UCHAR *)dst + 4;
							srcl[0] = (UCHAR)((5 * b[6] + b[5] + b[10] + b[9]) >> 3);
							srcl[1] = (UCHAR)((5 * g[6] + g[5] + g[10] + g[9]) >> 3);
							srcl[2] = (UCHAR)((5 * r[6] + r[5] + r[10] + r[9]) >> 3);

							srcl = (UCHAR *)dst + 8;
							srcl[0] = (UCHAR)((3 * b[7] + b[11]) >> 2);
							srcl[1] = (UCHAR)((3 * g[7] + g[11]) >> 2);
							srcl[2] = (UCHAR)((3 * r[7] + r[11]) >> 2);

							srcl = (UCHAR *)dst + fa->dst.pitch - 4;
							srcl[0] = (UCHAR)((3 * b[8] + b[4]) >> 2);
							srcl[1] = (UCHAR)((3 * g[8] + g[4]) >> 2);
							srcl[2] = (UCHAR)((3 * r[8] + r[4]) >> 2);

							srcl = (UCHAR *)dst + fa->dst.pitch;
							srcl[0] = (UCHAR)((5 * b[9] + b[10] + b[5] + b[6]) >> 3);
							srcl[1] = (UCHAR)((5 * g[9] + g[10] + g[5] + g[6]) >> 3);
							srcl[2] = (UCHAR)((5 * r[9] + r[10] + r[5] + r[6]) >> 3);

							srcl = (UCHAR *)dst + fa->dst.pitch + 4;
							srcl[0] = (UCHAR)((5 * b[10] + b[9] + b[6] + b[5]) >> 3);
							srcl[1] = (UCHAR)((5 * g[10] + g[9] + g[6] + g[5]) >> 3);
							srcl[2] = (UCHAR)((5 * r[10] + r[9] + r[6] + r[5]) >> 3);

							srcl = (UCHAR *)dst + fa->dst.pitch + 8;
							srcl[0] = (UCHAR)((3 * b[11] + b[7]) >> 2);
							srcl[1] = (UCHAR)((3 * g[11] + g[7]) >> 2);
							srcl[2] = (UCHAR)((3 * r[11] + r[7]) >> 2);

							srcl = (UCHAR *)dst + fa->dst.pitch * 2;
							srcl[0] = (UCHAR)((3 * b[13] + b[14]) >> 2);
							srcl[1] = (UCHAR)((3 * g[13] + g[14]) >> 2);
							srcl[2] = (UCHAR)((3 * r[13] + r[14]) >> 2);

							srcl = (UCHAR *)dst + fa->dst.pitch * 2 + 4;
							srcl[0] = (UCHAR)((3 * b[14] + b[13]) >> 2);
							srcl[1] = (UCHAR)((3 * g[14] + g[13]) >> 2);
							srcl[2] = (UCHAR)((3 * r[14] + r[13]) >> 2);

						}
						else if (borderv) {
							srcl = (UCHAR *)dst - fa->dst.pitch;
							srcl[0] = (UCHAR)((3 * b[1] + b[2]) >> 2);
							srcl[1] = (UCHAR)((3 * g[1] + g[2]) >> 2);
							srcl[2] = (UCHAR)((3 * r[1] + r[2]) >> 2);
							srcl = (UCHAR *)dst - fa->dst.pitch + 4;
							srcl[0] = (UCHAR)((3 * b[2] + b[1]) >> 2);
							srcl[1] = (UCHAR)((3 * g[2] + g[1]) >> 2);
							srcl[2] = (UCHAR)((3 * r[2] + r[1]) >> 2);

							srcl = (UCHAR *)dst;
							srcl[0] = (UCHAR)((3 * b[5] + b[6]) >> 2);
							srcl[1] = (UCHAR)((3 * g[5] + g[6]) >> 2);
							srcl[2] = (UCHAR)((3 * r[5] + r[6]) >> 2);
							srcl = (UCHAR *)dst + 4;
							srcl[0] = (UCHAR)((3 * b[6] + b[5]) >> 2);
							srcl[1] = (UCHAR)((3 * g[6] + g[5]) >> 2);
							srcl[2] = (UCHAR)((3 * r[6] + r[5]) >> 2);

							srcl = (UCHAR *)dst + fa->dst.pitch;
							srcl[0] = (UCHAR)((3 * b[9] + b[10]) >> 2);
							srcl[1] = (UCHAR)((3 * g[9] + g[10]) >> 2);
							srcl[2] = (UCHAR)((3 * r[9] + r[10]) >> 2);
							srcl = (UCHAR *)dst + fa->dst.pitch + 4;
							srcl[0] = (UCHAR)((3 * b[10] + b[9]) >> 2);
							srcl[1] = (UCHAR)((3 * g[10] + g[9]) >> 2);
							srcl[2] = (UCHAR)((3 * r[10] + r[9]) >> 2);

							srcl = (UCHAR *)dst + fa->dst.pitch * 2;
							srcl[0] = (UCHAR)((3 * b[13] + b[14]) >> 2);
							srcl[1] = (UCHAR)((3 * g[13] + g[14]) >> 2);
							srcl[2] = (UCHAR)((3 * r[13] + r[14]) >> 2);

							srcl = (UCHAR *)dst + fa->dst.pitch * 2 + 4;
							srcl[0] = (UCHAR)((3 * b[14] + b[13]) >> 2);
							srcl[1] = (UCHAR)((3 * g[14] + g[13]) >> 2);
							srcl[2] = (UCHAR)((3 * r[14] + r[13]) >> 2);
						}
						else if (borderh) {
							srcl = (UCHAR *)dst - 4;
							srcl[0] = (UCHAR)((3 * b[4] + b[8]) >> 2);
							srcl[1] = (UCHAR)((3 * g[4] + g[8]) >> 2);
							srcl[2] = (UCHAR)((3 * r[4] + r[8]) >> 2);
							srcl = (UCHAR *)dst + fa->dst.pitch - 4;
							srcl[0] = (UCHAR)((3 * b[8] + b[4]) >> 2);
							srcl[1] = (UCHAR)((3 * g[8] + g[4]) >> 2);
							srcl[2] = (UCHAR)((3 * r[8] + r[4]) >> 2);

							srcl = (UCHAR *)dst;
							srcl[0] = (UCHAR)((3 * b[5] + b[9]) >> 2);
							srcl[1] = (UCHAR)((3 * g[5] + g[9]) >> 2);
							srcl[2] = (UCHAR)((3 * r[5] + r[9]) >> 2);
							srcl = (UCHAR *)dst + fa->dst.pitch;
							srcl[0] = (UCHAR)((3 * b[9] + b[5]) >> 2);
							srcl[1] = (UCHAR)((3 * g[9] + g[5]) >> 2);
							srcl[2] = (UCHAR)((3 * r[9] + r[5]) >> 2);

							srcl = (UCHAR *)dst + 4;
							srcl[0] = (UCHAR)((3 * b[6] + b[10]) >> 2);
							srcl[1] = (UCHAR)((3 * g[6] + g[10]) >> 2);
							srcl[2] = (UCHAR)((3 * r[6] + r[10]) >> 2);
							srcl = (UCHAR *)dst + fa->dst.pitch + 4;
							srcl[0] = (UCHAR)((3 * b[10] + b[6]) >> 2);
							srcl[1] = (UCHAR)((3 * g[10] + g[6]) >> 2);
							srcl[2] = (UCHAR)((3 * r[10] + r[6]) >> 2);

							srcl = (UCHAR *)dst + 8;
							srcl[0] = (UCHAR)((3 * b[7] + b[11]) >> 2);
							srcl[1] = (UCHAR)((3 * g[7] + g[11]) >> 2);
							srcl[2] = (UCHAR)((3 * r[7] + r[11]) >> 2);
							srcl = (UCHAR *)dst + fa->dst.pitch + 8;
							srcl[0] = (UCHAR)((3 * b[11] + b[7]) >> 2);
							srcl[1] = (UCHAR)((3 * g[11] + g[7]) >> 2);
							srcl[2] = (UCHAR)((3 * r[11] + r[7]) >> 2);
						}	
					}							
				}
			dst++;
			nwpic++;
		}
		dst	= (Pixel32 *)((char *)dst + mfd->dalign);
	}

	if (!mfd->cascad || !mfd->dbldif) goto nodbl;
	dst = fa->dst.data;
	nwpic = nwpics;
	for (y = 0; y < height; y++)
	{
		for (x = 0; x < width; x++)
		{
			nwpic[0] = dst[0];
			dst++;
			nwpic++;
		}
		dst	= (Pixel32 *)((char *)dst + mfd->dalign);
	}
	dst = fa->dst.data;
	dst	= (Pixel32 *)((char *)dst + fa->dst.pitch + 4);
	nwpic = nwpics + mfd->width + 1;
	for (y = 1; y < (height - 2); y++)
	{
		for (x = 1; x < (width - 2); x++)
		{
			if (mfd->showpt && !mfd->cascad) {
				nnn = nwpic;
				pwid = mfd->width * 4;
			}
			else {
				nnn = dst;
				pwid = fa->dst.pitch;
			}
				srcl = (UCHAR *)nwpic - pwid - 4; 
				b[0] = srcl[0];
				g[0] = srcl[1];
				r[0] = srcl[2];
				srcl = (UCHAR *)nwpic - pwid;
				b[1] = srcl[0];
				g[1] = srcl[1];
				r[1] = srcl[2];
				srcl = (UCHAR *)nwpic - pwid + 4;
				b[2] = srcl[0];
				g[2] = srcl[1];
				r[2] = srcl[2];
				srcl = (UCHAR *)nwpic - pwid + 8;
				b[3] = srcl[0];
				g[3] = srcl[1];
				r[3] = srcl[2];
				srcl = (UCHAR *)nwpic - 4;
				b[4] = srcl[0];
				g[4] = srcl[1];
				r[4] = srcl[2];
				srcl = (UCHAR *)nwpic;
				b[5] = srcl[0];
				g[5] = srcl[1];
				r[5] = srcl[2];
				srcl = (UCHAR *)nwpic + 4;
				b[6] = srcl[0];
				g[6] = srcl[1];
				r[6] = srcl[2];
				srcl = (UCHAR *)nwpic + 8;
				b[7] = srcl[0];
				g[7] = srcl[1];
				r[7] = srcl[2];
				srcl = (UCHAR *)nwpic + pwid - 4;
				b[8] = srcl[0];
				g[8] = srcl[1];
				r[8] = srcl[2];
				srcl = (UCHAR *)nwpic + pwid;
				b[9] = srcl[0];
				g[9] = srcl[1];
				r[9] = srcl[2];
				srcl = (UCHAR *)nwpic + pwid + 4;
				b[10] = srcl[0];
				g[10] = srcl[1];
				r[10] = srcl[2];
				srcl = (UCHAR *)nwpic + pwid + 8;
				b[11] = srcl[0];
				g[11] = srcl[1];
				r[11] = srcl[2];
				srcl = (UCHAR *)nwpic + pwid * 2 - 4;
				b[12] = srcl[0];
				g[12] = srcl[1];
				r[12] = srcl[2];
				srcl = (UCHAR *)nwpic + pwid * 2;
				b[13] = srcl[0];
				g[13] = srcl[1];
				r[13] = srcl[2];
				srcl = (UCHAR *)nwpic + pwid * 2 + 4;
				b[14] = srcl[0];
				g[14] = srcl[1];
				r[14] = srcl[2];
				srcl = (UCHAR *)nwpic + pwid * 2 + 8;
				b[15] = srcl[0];
				g[15] = srcl[1];
				r[15] = srcl[2];

				dcv[0] = abs(r[1] - r[5]) +  abs(g[1] - g[5]) + abs(b[1] - b[5]);
				dcv[1] = abs(r[2] - r[6]) +  abs(g[2] - g[6]) + abs(b[2] - b[6]);
				dcv[2] = abs(r[9] - r[13]) +  abs(g[9] - g[13]) + abs(b[9] - b[13]);
				dcv[3] = abs(r[10] - r[14]) +  abs(g[10] - g[14]) + abs(b[10] - b[14]);
				dch[0] = abs(r[1] - r[2]) +  abs(g[1] - g[2]) + abs(b[1] - b[2]);
				dch[1] = abs(r[5] - r[6]) +  abs(g[5] - g[6]) + abs(b[5] - b[6]);
				dch[2] = abs(r[9] - r[10]) +  abs(g[9] - g[10]) + abs(b[9] - b[10]);
				dch[3] = abs(r[13] - r[14]) +  abs(g[13] - g[14]) + abs(b[13] - b[14]);
				
				long contr = 100 + mfd->dbhdam;
				bool borderv = 
					  (dcv[0] < mfd->dbldif)
					&& (dcv[1] < mfd->dbldif)
					&& (dcv[2] < mfd->dbldif)
					&& (dcv[3] < mfd->dbldif)
					&& (dch[0] > mfd->dbhdif)
					&& (dch[1] > mfd->dbhdif)
					&& (dch[2] > mfd->dbhdif)
					&& (dch[3] > mfd->dbhdif)
					&& (dch[0] < contr)
					&& (dch[1] < contr)
					&& (dch[2] < contr)
					&& (dch[3] < contr);

				dcv[0] = abs(r[4] - r[5]) +  abs(g[4] - g[5]) + abs(b[4] - b[5]);
				dcv[1] = abs(r[6] - r[7]) +  abs(g[6] - g[7]) + abs(b[6] - b[7]);
				dcv[2] = abs(r[8] - r[9]) +  abs(g[8] - g[9]) + abs(b[8] - b[9]);
				dcv[3] = abs(r[10] - r[11]) +  abs(g[10] - g[11]) + abs(b[10] - b[11]);
				dch[0] = abs(r[4] - r[8]) +  abs(g[4] - g[8]) + abs(b[4] - b[8]);
				dch[1] = abs(r[5] - r[9]) +  abs(g[5] - g[9]) + abs(b[5] - b[9]);
				dch[2] = abs(r[6] - r[10]) +  abs(g[6] - g[10]) + abs(b[6] - b[10]);
				dch[3] = abs(r[7] - r[11]) +  abs(g[7] - g[11]) + abs(b[7] - b[11]);

				bool borderh = 
					  ((dcv[0] < mfd->dbldif)
					&& (dcv[1] < mfd->dbldif)
					&& (dcv[2] < mfd->dbldif)
					&& (dcv[3] < mfd->dbldif)
					&& (dch[0] > mfd->dbhdif)
					&& (dch[1] > mfd->dbhdif)
					&& (dch[2] > mfd->dbhdif)
					&& (dch[3] > mfd->dbhdif)
					&& (dch[0] < contr)
					&& (dch[1] < contr)
					&& (dch[2] < contr)
					&& (dch[3] < contr));

				if (borderv || borderh) {
					if (mfd->showpt) {
						Pixel32 *lsrc;
						if (borderv && borderh) {
							lsrc = (Pixel32*)((UCHAR*)dst - fa->dst.pitch);
							lsrc[0] = 0xffff; 
							dst[0] = 0xffff;
							lsrc = (Pixel32*)((UCHAR*)dst + fa->dst.pitch);
							lsrc[0] = 0xffff; 
							lsrc = (Pixel32*)((UCHAR*)dst + fa->dst.pitch * 2);
							lsrc[0] = 0xffff;
							lsrc = (Pixel32*)((UCHAR*)dst - 4);
							lsrc[0] = 0xffff; 
							lsrc = (Pixel32*)((UCHAR*)dst + 4);
							lsrc[0] = 0xffff; 
							lsrc = (Pixel32*)((UCHAR*)dst + 8);
							lsrc[0] = 0xffff;
						}
						else if (borderv) {
							lsrc = (Pixel32*)((UCHAR*)dst - fa->dst.pitch);
							lsrc[0] = 0xffff; 
							dst[0] = 0xffff;
							lsrc = (Pixel32*)((UCHAR*)dst + fa->dst.pitch);
							lsrc[0] = 0xffff; 
							lsrc = (Pixel32*)((UCHAR*)dst + fa->dst.pitch * 2);
							lsrc[0] = 0xffff;
						}
						else if (borderh) {
							dst[0] = 0xffff;
							lsrc = (Pixel32*)((UCHAR*)dst - 4);
							lsrc[0] = 0xffff; 
							lsrc = (Pixel32*)((UCHAR*)dst + 4);
							lsrc[0] = 0xffff; 
							lsrc = (Pixel32*)((UCHAR*)dst + 8);
							lsrc[0] = 0xffff;
						}
					}
					else {
						if (borderv && borderh) {
							srcl = (UCHAR *)dst - fa->dst.pitch;
							srcl[0] = (UCHAR)((3 * b[1] + b[2]) >> 2);
							srcl[1] = (UCHAR)((3 * g[1] + g[2]) >> 2);
							srcl[2] = (UCHAR)((3 * r[1] + r[2]) >> 2);

							srcl = (UCHAR *)dst - fa->dst.pitch + 4;
							srcl[0] = (UCHAR)((b[2] + 3 * b[1]) >> 2);
							srcl[1] = (UCHAR)((g[2] + 3 * g[1]) >> 2);
							srcl[2] = (UCHAR)((r[2] + 3 * r[1]) >> 2);

							srcl = (UCHAR *)dst - 4;
							srcl[0] = (UCHAR)((3 * b[4] + b[8]) >> 2);
							srcl[1] = (UCHAR)((3 * g[4] + g[8]) >> 2);
							srcl[2] = (UCHAR)((3 * r[4] + r[8]) >> 2);

							srcl = (UCHAR *)dst;
							srcl[0] = (UCHAR)((5 * b[5] + b[6] + b[9] + b[10]) >> 3);
							srcl[1] = (UCHAR)((5 * g[5] + g[6] + g[9] + g[10]) >> 3);
							srcl[2] = (UCHAR)((5 * r[5] + r[6] + r[9] + r[10]) >> 3);

							srcl = (UCHAR *)dst + 4;
							srcl[0] = (UCHAR)((5 * b[6] + b[5] + b[10] + b[9]) >> 3);
							srcl[1] = (UCHAR)((5 * g[6] + g[5] + g[10] + g[9]) >> 3);
							srcl[2] = (UCHAR)((5 * r[6] + r[5] + r[10] + r[9]) >> 3);

							srcl = (UCHAR *)dst + 8;
							srcl[0] = (UCHAR)((3 * b[7] + b[11]) >> 2);
							srcl[1] = (UCHAR)((3 * g[7] + g[11]) >> 2);
							srcl[2] = (UCHAR)((3 * r[7] + r[11]) >> 2);

							srcl = (UCHAR *)dst + fa->dst.pitch - 4;
							srcl[0] = (UCHAR)((3 * b[8] + b[4]) >> 2);
							srcl[1] = (UCHAR)((3 * g[8] + g[4]) >> 2);
							srcl[2] = (UCHAR)((3 * r[8] + r[4]) >> 2);

							srcl = (UCHAR *)dst + fa->dst.pitch;
							srcl[0] = (UCHAR)((5 * b[9] + b[10] + b[5] + b[6]) >> 3);
							srcl[1] = (UCHAR)((5 * g[9] + g[10] + g[5] + g[6]) >> 3);
							srcl[2] = (UCHAR)((5 * r[9] + r[10] + r[5] + r[6]) >> 3);

							srcl = (UCHAR *)dst + fa->dst.pitch + 4;
							srcl[0] = (UCHAR)((5 * b[10] + b[9] + b[6] + b[5]) >> 3);
							srcl[1] = (UCHAR)((5 * g[10] + g[9] + g[6] + g[5]) >> 3);
							srcl[2] = (UCHAR)((5 * r[10] + r[9] + r[6] + r[5]) >> 3);

							srcl = (UCHAR *)dst + fa->dst.pitch + 8;
							srcl[0] = (UCHAR)((3 * b[11] + b[7]) >> 2);
							srcl[1] = (UCHAR)((3 * g[11] + g[7]) >> 2);
							srcl[2] = (UCHAR)((3 * r[11] + r[7]) >> 2);

							srcl = (UCHAR *)dst + fa->dst.pitch * 2;
							srcl[0] = (UCHAR)((3 * b[13] + b[14]) >> 2);
							srcl[1] = (UCHAR)((3 * g[13] + g[14]) >> 2);
							srcl[2] = (UCHAR)((3 * r[13] + r[14]) >> 2);

							srcl = (UCHAR *)dst + fa->dst.pitch * 2 + 4;
							srcl[0] = (UCHAR)((3 * b[14] + b[13]) >> 2);
							srcl[1] = (UCHAR)((3 * g[14] + g[13]) >> 2);
							srcl[2] = (UCHAR)((3 * r[14] + r[13]) >> 2);

						}
						else if (borderv) {
							srcl = (UCHAR *)dst - fa->dst.pitch;
							srcl[0] = (UCHAR)((3 * b[1] + b[2]) >> 2);
							srcl[1] = (UCHAR)((3 * g[1] + g[2]) >> 2);
							srcl[2] = (UCHAR)((3 * r[1] + r[2]) >> 2);
							srcl = (UCHAR *)dst - fa->dst.pitch + 4;
							srcl[0] = (UCHAR)((3 * b[2] + b[1]) >> 2);
							srcl[1] = (UCHAR)((3 * g[2] + g[1]) >> 2);
							srcl[2] = (UCHAR)((3 * r[2] + r[1]) >> 2);

							srcl = (UCHAR *)dst;
							srcl[0] = (UCHAR)((3 * b[5] + b[6]) >> 2);
							srcl[1] = (UCHAR)((3 * g[5] + g[6]) >> 2);
							srcl[2] = (UCHAR)((3 * r[5] + r[6]) >> 2);
							srcl = (UCHAR *)dst + 4;
							srcl[0] = (UCHAR)((3 * b[6] + b[5]) >> 2);
							srcl[1] = (UCHAR)((3 * g[6] + g[5]) >> 2);
							srcl[2] = (UCHAR)((3 * r[6] + r[5]) >> 2);

							srcl = (UCHAR *)dst + fa->dst.pitch;
							srcl[0] = (UCHAR)((3 * b[9] + b[10]) >> 2);
							srcl[1] = (UCHAR)((3 * g[9] + g[10]) >> 2);
							srcl[2] = (UCHAR)((3 * r[9] + r[10]) >> 2);
							srcl = (UCHAR *)dst + fa->dst.pitch + 4;
							srcl[0] = (UCHAR)((3 * b[10] + b[9]) >> 2);
							srcl[1] = (UCHAR)((3 * g[10] + g[9]) >> 2);
							srcl[2] = (UCHAR)((3 * r[10] + r[9]) >> 2);

							srcl = (UCHAR *)dst + fa->dst.pitch * 2;
							srcl[0] = (UCHAR)((3 * b[13] + b[14]) >> 2);
							srcl[1] = (UCHAR)((3 * g[13] + g[14]) >> 2);
							srcl[2] = (UCHAR)((3 * r[13] + r[14]) >> 2);

							srcl = (UCHAR *)dst + fa->dst.pitch * 2 + 4;
							srcl[0] = (UCHAR)((3 * b[14] + b[13]) >> 2);
							srcl[1] = (UCHAR)((3 * g[14] + g[13]) >> 2);
							srcl[2] = (UCHAR)((3 * r[14] + r[13]) >> 2);
						}
						else if (borderh) {
							srcl = (UCHAR *)dst - 4;
							srcl[0] = (UCHAR)((3 * b[4] + b[8]) >> 2);
							srcl[1] = (UCHAR)((3 * g[4] + g[8]) >> 2);
							srcl[2] = (UCHAR)((3 * r[4] + r[8]) >> 2);
							srcl = (UCHAR *)dst + fa->dst.pitch - 4;
							srcl[0] = (UCHAR)((3 * b[8] + b[4]) >> 2);
							srcl[1] = (UCHAR)((3 * g[8] + g[4]) >> 2);
							srcl[2] = (UCHAR)((3 * r[8] + r[4]) >> 2);

							srcl = (UCHAR *)dst;
							srcl[0] = (UCHAR)((3 * b[5] + b[9]) >> 2);
							srcl[1] = (UCHAR)((3 * g[5] + g[9]) >> 2);
							srcl[2] = (UCHAR)((3 * r[5] + r[9]) >> 2);
							srcl = (UCHAR *)dst + fa->dst.pitch;
							srcl[0] = (UCHAR)((3 * b[9] + b[5]) >> 2);
							srcl[1] = (UCHAR)((3 * g[9] + g[5]) >> 2);
							srcl[2] = (UCHAR)((3 * r[9] + r[5]) >> 2);

							srcl = (UCHAR *)dst + 4;
							srcl[0] = (UCHAR)((3 * b[6] + b[10]) >> 2);
							srcl[1] = (UCHAR)((3 * g[6] + g[10]) >> 2);
							srcl[2] = (UCHAR)((3 * r[6] + r[10]) >> 2);
							srcl = (UCHAR *)dst + fa->dst.pitch + 4;
							srcl[0] = (UCHAR)((3 * b[10] + b[6]) >> 2);
							srcl[1] = (UCHAR)((3 * g[10] + g[6]) >> 2);
							srcl[2] = (UCHAR)((3 * r[10] + r[6]) >> 2);

							srcl = (UCHAR *)dst + 8;
							srcl[0] = (UCHAR)((3 * b[7] + b[11]) >> 2);
							srcl[1] = (UCHAR)((3 * g[7] + g[11]) >> 2);
							srcl[2] = (UCHAR)((3 * r[7] + r[11]) >> 2);
							srcl = (UCHAR *)dst + fa->dst.pitch + 8;
							srcl[0] = (UCHAR)((3 * b[11] + b[7]) >> 2);
							srcl[1] = (UCHAR)((3 * g[11] + g[7]) >> 2);
							srcl[2] = (UCHAR)((3 * r[11] + r[7]) >> 2);
						}	
					}							
				}
			dst++;
			nwpic++;
		}
		dst	= (Pixel32 *)((char *)dst + mfd->dalign);
	}

nodbl:
	delete nwpics;
	return 0;
}

extern "C" int __declspec(dllexport) __cdecl VirtualdubFilterModuleInit2(FilterModule *fm, const FilterFunctions *ff, int& vdfd_ver, int& vdfd_compat);
extern "C" void __declspec(dllexport) __cdecl VirtualdubFilterModuleDeinit(FilterModule *fm, const FilterFunctions *ff);

static FilterDefinition *fd_tutorial;

int __declspec(dllexport) __cdecl VirtualdubFilterModuleInit2(FilterModule *fm, const FilterFunctions *ff, int& vdfd_ver, int& vdfd_compat) {
	if (!(fd_tutorial = ff->addFilter(fm, &filterDef_tutorial, sizeof(FilterDefinition))))
		return 1;

	vdfd_ver = VIRTUALDUB_FILTERDEF_VERSION;
	vdfd_compat = VIRTUALDUB_FILTERDEF_COMPATIBLE;

	return 0;
}

void __declspec(dllexport) __cdecl VirtualdubFilterModuleDeinit(FilterModule *fm, const FilterFunctions *ff) {
	ff->removeFilter(fd_tutorial);
}

int InitProc(FilterActivation *fa, const FilterFunctions *ff) {
	MyFilterData *mfd = (MyFilterData *)fa->filter_data;

	mfd->rshpt			= 0;
	mfd->rshxx			= 0;
	mfd->rshyy			= 0;
	mfd->gshpt			= 0;
	mfd->gshxx			= 0;
	mfd->gshyy			= 0;
	mfd->bshpt			= 0;
	mfd->bshxx			= 0;
	mfd->bshyy			= 0;
	mfd->dbldif			= 0;
	mfd->dbhdif			= 0;
	mfd->dbhdam			= 0;

	mfd->showpt		= FALSE;
	mfd->cascad		= FALSE;
	mfd->dbllock		= FALSE;
	mfd->keeppt			= TRUE;
	mfd->rshlock		= FALSE;
	mfd->gshlock		= FALSE;
	mfd->bshlock		= FALSE;
	mfd->shiunion		= TRUE;
	mfd->nmfunc			= IDC_RSH;
	return 0;
}

BOOL CALLBACK ConfigDlgProc(HWND hdlg, UINT msg, WPARAM wParam, LPARAM lParam)
{
	MyFilterData *mfd = (MyFilterData *)GetWindowLong(hdlg, DWL_USER);
	int mPr=0;
	bool redo = false;
	bool undo = false;

	switch(msg) {
		case WM_INITDIALOG:
			SetWindowLong(hdlg, DWL_USER, lParam);
			mfd = (MyFilterData *)lParam;
			HWND hWnd;

			setFlg(mfd);
			hWnd = GetDlgItem(hdlg, IDC_EF1);
			SendMessage(hWnd, TBM_SETRANGE, (WPARAM)TRUE, MAKELONG(-100, 100));
			SendMessage(hWnd, TBM_SETTICFREQ, 10, 0);
			hWnd = GetDlgItem(hdlg, IDC_EF2);
			SendMessage(hWnd, TBM_SETRANGE, (WPARAM)TRUE, MAKELONG(-100, 100));
			SendMessage(hWnd, TBM_SETTICFREQ, 10, 0);
			hWnd = GetDlgItem(hdlg, IDC_EF3);
			SendMessage(hWnd, TBM_SETRANGE, (WPARAM)TRUE, MAKELONG(-100, 100));
			SendMessage(hWnd, TBM_SETTICFREQ, 10, 0);

			SendMessage(GetDlgItem(hdlg, IDC_KEEPPT), BM_SETCHECK, mfd->keeppt, 0);
			SendMessage(GetDlgItem(hdlg, IDC_SHIUNION), BM_SETCHECK, mfd->shiunion, 0);
			SendMessage(GetDlgItem(hdlg, IDC_RSHLOCK), BM_SETCHECK, mfd->rshlock, 0);
			SendMessage(GetDlgItem(hdlg, IDC_GSHLOCK), BM_SETCHECK, mfd->gshlock, 0);
			SendMessage(GetDlgItem(hdlg, IDC_BSHLOCK), BM_SETCHECK, mfd->bshlock, 0);
			SendMessage(GetDlgItem(hdlg, IDC_SHOWPT), BM_SETCHECK, mfd->showpt, 0);
			SendMessage(GetDlgItem(hdlg, IDC_CASCAD), BM_SETCHECK, mfd->cascad, 0);
			SendMessage(GetDlgItem(hdlg, IDC_DBRLOCK), BM_SETCHECK, mfd->dbllock, 0);

			SendMessage(GetDlgItem(hdlg, mfd->nmfunc), BM_CLICK, TRUE, 0);


			mfd->ifp->InitButton(GetDlgItem(hdlg, IDPREVIEW));
			mfd->ifp->Toggle(hdlg);
			return TRUE;

		case WM_COMMAND:
			switch(LOWORD(wParam)) {
			case IDPREVIEW:
				mfd->ifp->Toggle(hdlg);
				break;
			case IDOK:
				EndDialog(hdlg, 0);
				return TRUE;
			case IDHELP:
				{
				char prog[256];
				char path[256];
				LPTSTR ptr;
				GetModuleFileName(NULL, prog, 255);
				GetFullPathName(prog, 255, path, &ptr);
				*ptr = 0;
				strcat(path, "plugins\\DE.txt");
				OutputDebugString(path);
				OutputDebugString("\n");
				strcpy(prog, "Notepad ");
				strcat(prog, path);
				WinExec(prog, SW_SHOW);
				return TRUE;
				}
			case IDCANCEL:
				EndDialog(hdlg, 1);
				return TRUE;
			case IDC_ZEF1:
				switch (mfd->nmfunc) {
				case IDC_RSH:
					if (mfd->rshlock) undo = true;
					else if (mfd->shiunion) {
							mfd->rshpt = 0;
							mfd->gshpt = 0;
							mfd->bshpt = 0;
					}
					else mfd->rshpt = 0;
					break;
				case IDC_GSH:
					if (mfd->gshlock) undo = true;
					else if (mfd->shiunion) {
							mfd->rshpt = 0;
							mfd->gshpt = 0;
							mfd->bshpt = 0;
					}
					else mfd->gshpt = 0;
					break;
				case IDC_BSH:
					if (mfd->bshlock) undo = true;
					else if (mfd->shiunion) {
							mfd->rshpt = 0;
							mfd->gshpt = 0;
							mfd->bshpt = 0;
					}
					else mfd->bshpt = 0;
					break;
				case IDC_DBR:
					if (mfd->dbllock) undo = true;
					else mfd->dbldif = 0;
					break;
				}
				if (!undo) {
					SendMessage(GetDlgItem(hdlg, IDC_EF1), TBM_SETPOS, (WPARAM)TRUE, 0);
					SetDlgItemInt(hdlg, IDC_EF1VAL, 0, TRUE);
					setFlg(mfd);
					mfd->ifp->RedoFrame();
				}
				undo = false;
				return TRUE;
			case IDC_ZEF2:
				switch (mfd->nmfunc) {
				case IDC_RSH:
					if (mfd->rshlock) undo = true;
					else if (mfd->shiunion) {
							mfd->rshxx = 0;
							mfd->gshxx = 0;
							mfd->bshxx = 0;
					}
					else mfd->rshxx = 0;
					break;
				case IDC_GSH:
					if (mfd->gshlock) undo = true;
					else if (mfd->shiunion) {
							mfd->rshxx = 0;
							mfd->gshxx = 0;
							mfd->bshxx = 0;
					}
					else mfd->gshxx = 0;
					break;
				case IDC_BSH:
					if (mfd->bshlock) undo = true;
					else if (mfd->shiunion) {
							mfd->rshxx = 0;
							mfd->gshxx = 0;
							mfd->bshxx = 0;
					}
					else mfd->bshxx = 0;
					break;
				case IDC_DBR:
					if (mfd->dbllock) undo = true;
					else mfd->dbhdif = 0;
					break;
				}
				if ( !undo) {
					SendMessage(GetDlgItem(hdlg, IDC_EF2), TBM_SETPOS, (WPARAM)TRUE, 0);
					SetDlgItemInt(hdlg, IDC_EF2VAL, 0, TRUE);
					setFlg(mfd);
					mfd->ifp->RedoFrame();
				}
				undo = false;
				return TRUE;
			case IDC_ZEF3:
				switch (mfd->nmfunc) {
				case IDC_RSH:
					if (mfd->rshlock) undo = true;
					else if (mfd->shiunion) {
							mfd->rshyy = 0;
							mfd->gshyy = 0;
							mfd->bshyy = 0;
					}
					else mfd->rshyy = 0;
					break;
				case IDC_GSH:
					if (mfd->gshlock) undo = true;
					else if (mfd->shiunion) {
							mfd->rshyy = 0;
							mfd->gshyy = 0;
							mfd->bshyy = 0;
					}
					else mfd->gshyy = 0;
					break;
				case IDC_BSH:
					if (mfd->bshlock) undo = true;
					else if (mfd->shiunion) {
							mfd->rshyy = 0;
							mfd->gshyy = 0;
							mfd->bshyy = 0;
					}
					else mfd->bshyy = 0;
					break;
				case IDC_DBR:
					if (mfd->dbllock) undo = true;
					else mfd->dbhdam = 0;
					break;
				}
				if (!undo) {
				SendMessage(GetDlgItem(hdlg, IDC_EF3), TBM_SETPOS, (WPARAM)TRUE, 0);
				SetDlgItemInt(hdlg, IDC_EF3VAL, 0, TRUE);
				setFlg(mfd);
				mfd->ifp->RedoFrame();
				}
				undo = false;
				return TRUE;
			case IDC_ZEROALL:
				SendMessage(GetDlgItem(hdlg, IDC_RSHLOCK), BM_SETCHECK, FALSE, 0);
				SendMessage(GetDlgItem(hdlg, IDC_GSHLOCK), BM_SETCHECK, FALSE, 0);
				SendMessage(GetDlgItem(hdlg, IDC_BSHLOCK), BM_SETCHECK, FALSE, 0);
				SendMessage(GetDlgItem(hdlg, IDC_SHIUNION), BM_SETCHECK, TRUE, 0);
				SendMessage(GetDlgItem(hdlg, IDC_KEEPPT), BM_SETCHECK, TRUE, 0);
				SendMessage(GetDlgItem(hdlg, IDC_SHOWPT), BM_SETCHECK, FALSE, 0);
				SendMessage(GetDlgItem(hdlg, IDC_CASCAD), BM_SETCHECK, FALSE, 0);
				SendMessage(GetDlgItem(hdlg, IDC_DBRLOCK), BM_SETCHECK, FALSE, 0);
				mfd->rshlock = FALSE;
				mfd->gshlock = FALSE;
				mfd->bshlock = FALSE;
				mfd->shiunion = TRUE;
				mfd->keeppt = TRUE;
				mfd->showpt = FALSE;
				mfd->cascad = FALSE;
				mfd->dbllock= FALSE;

				SendMessage(GetDlgItem(hdlg, IDC_EF1), TBM_SETPOS, (WPARAM)TRUE, 0);
				SetDlgItemInt(hdlg, IDC_EF1VAL, 0, TRUE);
				mfd->rshpt = 0;
				mfd->gshpt = 0;
				mfd->bshpt = 0;
				mfd->dbldif = 0;
				SendMessage(GetDlgItem(hdlg, IDC_EF2), TBM_SETPOS, (WPARAM)TRUE, 0);
				SetDlgItemInt(hdlg, IDC_EF2VAL, 0, TRUE);
				mfd->rshxx = 0;
				mfd->gshxx = 0;
				mfd->bshxx = 0;
				mfd->dbhdif = 0;
				SendMessage(GetDlgItem(hdlg, IDC_EF3), TBM_SETPOS, (WPARAM)TRUE, 0);
				SetDlgItemInt(hdlg, IDC_EF3VAL, 0, TRUE);
				mfd->rshyy = 0;
				mfd->gshyy = 0;
				mfd->bshyy = 0;
				mfd->dbhdam = 0;
				setFlg(mfd);
				SetDlgItemText(hdlg, IDC_FUNAM, f1ntxt);
				mfd->ifp->RedoFrame();
				return TRUE;
			case IDC_RSH:
				ShowWindow(GetDlgItem(hdlg, IDC_KEEPPT), SW_SHOW);
				ShowWindow(GetDlgItem(hdlg, IDC_SHOWPT), SW_HIDE);
				ShowWindow(GetDlgItem(hdlg, IDC_CASCAD), SW_HIDE);
				SetDlgItemText(hdlg, IDC_EF1TEXT, f1ef1txt);
				SetDlgItemText(hdlg, IDC_EF2TEXT, f1ef2txt);
				SetDlgItemText(hdlg, IDC_EF3TEXT, f1ef3txt);
				if (mfd->shiunion) SetDlgItemText(hdlg, IDC_FUNAM, f1ntxt);
				else SetDlgItemText(hdlg, IDC_FUNAM, f11txt);

				SendMessage(GetDlgItem(hdlg, IDC_EF1), TBM_SETPOS, (WPARAM)TRUE, -mfd->rshpt);
				SetDlgItemInt(hdlg, IDC_EF1VAL, mfd->rshpt, TRUE);
				SendMessage(GetDlgItem(hdlg, IDC_EF2), TBM_SETPOS, (WPARAM)TRUE, -mfd->rshxx);
				SetDlgItemInt(hdlg, IDC_EF2VAL, mfd->rshxx, TRUE);
				SendMessage(GetDlgItem(hdlg, IDC_EF3), TBM_SETPOS, (WPARAM)TRUE, -mfd->rshyy);
				SetDlgItemInt(hdlg, IDC_EF3VAL, mfd->rshyy, TRUE);
				mfd->nmfunc = IDC_RSH;
				return TRUE;
			case IDC_GSH:
				ShowWindow(GetDlgItem(hdlg, IDC_KEEPPT), SW_SHOW);
				ShowWindow(GetDlgItem(hdlg, IDC_SHOWPT), SW_HIDE);
				ShowWindow(GetDlgItem(hdlg, IDC_CASCAD), SW_HIDE);
				SetDlgItemText(hdlg, IDC_EF1TEXT, f1ef1txt);
				SetDlgItemText(hdlg, IDC_EF2TEXT, f1ef2txt);
				SetDlgItemText(hdlg, IDC_EF3TEXT, f1ef3txt);
				if (mfd->shiunion) SetDlgItemText(hdlg, IDC_FUNAM, f1ntxt);
				else SetDlgItemText(hdlg, IDC_FUNAM, f12txt);

				SendMessage(GetDlgItem(hdlg, IDC_EF1), TBM_SETPOS, (WPARAM)TRUE, -mfd->gshpt);
				SetDlgItemInt(hdlg, IDC_EF1VAL, mfd->gshpt, TRUE);
				SendMessage(GetDlgItem(hdlg, IDC_EF2), TBM_SETPOS, (WPARAM)TRUE, -mfd->gshxx);
				SetDlgItemInt(hdlg, IDC_EF2VAL, mfd->gshxx, TRUE);
				SendMessage(GetDlgItem(hdlg, IDC_EF3), TBM_SETPOS, (WPARAM)TRUE, -mfd->gshyy);
				SetDlgItemInt(hdlg, IDC_EF3VAL, mfd->gshyy, TRUE);
				mfd->nmfunc = IDC_GSH;
				return TRUE;
			case IDC_BSH:
				ShowWindow(GetDlgItem(hdlg, IDC_KEEPPT), SW_SHOW);
				ShowWindow(GetDlgItem(hdlg, IDC_SHOWPT), SW_HIDE);
				ShowWindow(GetDlgItem(hdlg, IDC_CASCAD), SW_HIDE);
				SetDlgItemText(hdlg, IDC_EF1TEXT, f1ef1txt);
				SetDlgItemText(hdlg, IDC_EF2TEXT, f1ef2txt);
				SetDlgItemText(hdlg, IDC_EF3TEXT, f1ef3txt);
				if (mfd->shiunion) SetDlgItemText(hdlg, IDC_FUNAM, f1ntxt);
				else SetDlgItemText(hdlg, IDC_FUNAM, f13txt);
				
				SendMessage(GetDlgItem(hdlg, IDC_EF1), TBM_SETPOS, (WPARAM)TRUE, -mfd->bshpt);
				SetDlgItemInt(hdlg, IDC_EF1VAL, mfd->bshpt, TRUE);
				SendMessage(GetDlgItem(hdlg, IDC_EF2), TBM_SETPOS, (WPARAM)TRUE, -mfd->bshxx);
				SetDlgItemInt(hdlg, IDC_EF2VAL, mfd->bshxx, TRUE);
				SendMessage(GetDlgItem(hdlg, IDC_EF3), TBM_SETPOS, (WPARAM)TRUE, -mfd->bshyy);
				SetDlgItemInt(hdlg, IDC_EF3VAL, mfd->bshyy, TRUE);
				mfd->nmfunc = IDC_BSH;
				return TRUE;
			case IDC_DBR:
				ShowWindow(GetDlgItem(hdlg, IDC_KEEPPT), SW_HIDE);
				ShowWindow(GetDlgItem(hdlg, IDC_SHOWPT), SW_SHOW);
				ShowWindow(GetDlgItem(hdlg, IDC_CASCAD), SW_SHOW);
				SetDlgItemText(hdlg, IDC_EF1TEXT, f2ef1txt);
				SetDlgItemText(hdlg, IDC_EF2TEXT, f2ef2txt);
				SetDlgItemText(hdlg, IDC_EF3TEXT, f2ef3txt);
				SetDlgItemText(hdlg, IDC_FUNAM, f2txt);
				
				SendMessage(GetDlgItem(hdlg, IDC_EF1), TBM_SETPOS, (WPARAM)TRUE, -mfd->dbldif);
				SetDlgItemInt(hdlg, IDC_EF1VAL, mfd->dbldif, TRUE);
				SendMessage(GetDlgItem(hdlg, IDC_EF2), TBM_SETPOS, (WPARAM)TRUE, -mfd->dbhdif);
				SetDlgItemInt(hdlg, IDC_EF2VAL, mfd->dbhdif, TRUE);
				SendMessage(GetDlgItem(hdlg, IDC_EF3), TBM_SETPOS, (WPARAM)TRUE, -mfd->dbhdam);
				SetDlgItemInt(hdlg, IDC_EF3VAL, mfd->dbhdam, TRUE);
				mfd->nmfunc = IDC_DBR;
				return TRUE;
			case IDC_KEEPPT: {
				mfd->keeppt = IsDlgButtonChecked(hdlg, IDC_KEEPPT);
				setFlg(mfd);
				mfd->ifp->RedoFrame();}
				return TRUE;
			case IDC_SHOWPT: {
				mfd->showpt = IsDlgButtonChecked(hdlg, IDC_SHOWPT);
				setFlg(mfd);
				mfd->ifp->RedoFrame();}
				return TRUE;
			case IDC_CASCAD: {
				mfd->cascad = IsDlgButtonChecked(hdlg, IDC_CASCAD);
				setFlg(mfd);
				mfd->ifp->RedoFrame();}
				return TRUE;
			case IDC_RSHLOCK:
			case IDC_GSHLOCK:
			case IDC_BSHLOCK:
				if (IsDlgButtonChecked(hdlg, LOWORD(wParam))) {
				SendMessage(GetDlgItem(hdlg, IDC_SHIUNION), BM_SETCHECK, FALSE, 0);
				mfd->shiunion = FALSE;
				}
				if (mfd->shiunion) SetDlgItemText(hdlg, IDC_FUNAM, f1ntxt);
				else {
					switch (mfd->nmfunc) {
						case IDC_RSH:
							SetDlgItemText(hdlg, IDC_FUNAM, f11txt);
							break;
						case IDC_GSH:
							SetDlgItemText(hdlg, IDC_FUNAM, f12txt);
							break;
						case IDC_BSH:
							SetDlgItemText(hdlg, IDC_FUNAM, f13txt);
							break;
					}
				}
				mfd->rshlock = IsDlgButtonChecked(hdlg, IDC_RSHLOCK);
				mfd->gshlock = IsDlgButtonChecked(hdlg, IDC_GSHLOCK);
				mfd->bshlock = IsDlgButtonChecked(hdlg, IDC_BSHLOCK);
				return TRUE;
			case IDC_DBRLOCK:
				mfd->dbllock = IsDlgButtonChecked(hdlg, IDC_DBRLOCK);
				return TRUE;
			case IDC_SHIUNION:
				if (mfd->rshlock || mfd->gshlock || mfd->bshlock) {
					SendMessage(GetDlgItem(hdlg, IDC_SHIUNION), BM_SETCHECK, FALSE, 0);
					mfd->shiunion = FALSE;
				}
				else mfd->shiunion = IsDlgButtonChecked(hdlg, IDC_SHIUNION);
				if (mfd->shiunion) SetDlgItemText(hdlg, IDC_FUNAM, f1ntxt);
				else {
					switch (mfd->nmfunc) {
						case IDC_RSH:
							SetDlgItemText(hdlg, IDC_FUNAM, f11txt);
							break;
						case IDC_GSH:
							SetDlgItemText(hdlg, IDC_FUNAM, f12txt);
							break;
						case IDC_BSH:
							SetDlgItemText(hdlg, IDC_FUNAM, f13txt);
							break;
					}
				}
				return TRUE;
			}
			break;
		case WM_VSCROLL:
			{
				int ef1, ef2, ef3, ef1old, ef2old, ef3old, dpos, dw, dh;
				ef1 = -SendMessage(GetDlgItem(hdlg, IDC_EF1), TBM_GETPOS, 0, 0);
				ef2 = -SendMessage(GetDlgItem(hdlg, IDC_EF2), TBM_GETPOS, 0, 0);
				ef3 = -SendMessage(GetDlgItem(hdlg, IDC_EF3), TBM_GETPOS, 0, 0);
				if (mfd->width > 100) dw = 100;
				else dw = mfd->width;
				if (mfd->height > 100) dh = 100;
				else dh = mfd->height;
				switch (mfd->nmfunc) {
				case IDC_RSH:
						ef1old = mfd->rshpt;
						ef2old = mfd->rshxx;
						ef3old = mfd->rshyy;
					if (mfd->rshlock) undo = true;
					else {
						if((mfd->rshpt != ef1) || (mfd->rshxx != ef2) || (mfd->rshyy != ef3)) {
							redo = true;
							if (mfd->shiunion) {
								if (ef1 != ef1old) {
									dpos = ef1 - ef1old;
									dpos = normpos(mfd->rshpt, mfd->gshpt, mfd->bshpt, dpos, dw);
									ef1 = ef1old + dpos;
									mfd->gshpt += dpos;
									mfd->bshpt += dpos;
								}
								if (ef2 != ef2old) {
									dpos = ef2 - ef2old;
									dpos = normpos(mfd->rshxx, mfd->gshxx, mfd->bshxx, dpos, dw);
									ef2 = ef2old + dpos;
									mfd->gshxx += dpos;
									mfd->bshxx += dpos;
								}
								if (ef3 != ef3old) {
									dpos = ef3 - ef3old;
									dpos = normpos(mfd->rshyy, mfd->gshyy, mfd->bshyy, dpos, dh);
									ef3 = ef3old + dpos;
									mfd->gshyy += dpos;
									mfd->bshyy += dpos;
								}
							}
							mfd->rshpt = normv(ef1, dw);
							mfd->rshxx = normv(ef2, dw);
							mfd->rshyy = normv(ef3, dh);
							ef1 = mfd->rshpt;
							ef2 = mfd->rshxx;
							ef3 = mfd->rshyy;
						}
					}
					break;
				case IDC_GSH:
						ef1old = mfd->gshpt;
						ef2old = mfd->gshxx;
						ef3old = mfd->gshyy;
					if (mfd->gshlock) undo = true;
					else {
						if((mfd->gshpt != ef1) || (mfd->gshxx != ef2) || (mfd->gshyy != ef3)) {
							redo = true;
							if (mfd->shiunion) {
								if (ef1 != ef1old) {
									dpos = ef1 - ef1old;
									dpos = normpos(mfd->rshpt, mfd->gshpt, mfd->bshpt, dpos, dw);
									ef1 = ef1old + dpos;
									mfd->rshpt += dpos;
									mfd->bshpt += dpos;
								}
								if (ef2 != ef2old) {
									dpos = ef2 - ef2old;
									dpos = normpos(mfd->rshxx, mfd->gshxx, mfd->bshxx, dpos, dw);
									ef2 = ef2old + dpos;
									mfd->rshxx += dpos;
									mfd->bshxx += dpos;
								}
								if (ef3 != ef3old) {
									dpos = ef3 - ef3old;
									dpos = normpos(mfd->rshyy, mfd->gshyy, mfd->bshyy, dpos, dh);
									ef3 = ef3old + dpos;
									mfd->rshyy += dpos;
									mfd->bshyy += dpos;
								}
							}
							mfd->gshpt = normv(ef1, dw);
							mfd->gshxx = normv(ef2, dw);
							mfd->gshyy = normv(ef3, dh);
							ef1 = mfd->gshpt;
							ef2 = mfd->gshxx;
							ef3 = mfd->gshyy;
						}
					}
					break;
				case IDC_BSH:
						ef1old = mfd->bshpt;
						ef2old = mfd->bshxx;
						ef3old = mfd->bshyy;
					if (mfd->bshlock) undo = true;
					else {
						if((mfd->bshpt != ef1) || (mfd->bshxx != ef2) || (mfd->bshyy != ef3)) {
							redo = true;
							if (mfd->shiunion) {
								if (ef1 != ef1old) {
									dpos = ef1 - ef1old;
									dpos = normpos(mfd->rshpt, mfd->gshpt, mfd->bshpt, dpos, dw);
									ef1 = ef1old + dpos;
									mfd->rshpt += dpos;
									mfd->gshpt += dpos;
								}
								if (ef2 != ef2old) {
									dpos = ef2 - ef2old;
									dpos = normpos(mfd->rshxx, mfd->gshxx, mfd->bshxx, dpos, dw);
									ef2 = ef2old + dpos;
									mfd->rshxx += dpos;
									mfd->gshxx += dpos;
								}
								if (ef3 != ef3old) {
									dpos = ef3 - ef3old;
									dpos = normpos(mfd->rshyy, mfd->gshyy, mfd->bshyy, dpos, dh);
									ef3 = ef3old + dpos;
									mfd->rshyy += dpos;
									mfd->gshyy += dpos;
								}
							}
							mfd->bshpt = normv(ef1, dw);
							mfd->bshxx = normv(ef2, dw);
							mfd->bshyy = normv(ef3, dh);
							ef1 = mfd->bshpt;
							ef2 = mfd->bshxx;
							ef3 = mfd->bshyy;
						}
					}
					break;
				case IDC_DBR:
						ef1old = mfd->dbldif;
						ef2old = mfd->dbhdif;
						ef3old = mfd->dbhdam;
					if (mfd->dbllock) undo = true;
					else {
						if((mfd->dbldif != ef1) || (mfd->dbhdif != ef2) || (mfd->dbhdam != ef3)) {
							redo = true;
							if (ef1 < 0) mfd->dbldif = 0;
							else mfd->dbldif = ef1;
							ef1 = mfd->dbldif;
							if (ef2 < -1) mfd->dbhdif = -1;
							else mfd->dbhdif = ef2;
							ef2 = mfd->dbhdif;
							mfd->dbhdam = ef3;
						}
					}
					break;
				}
				if (undo) {
					SendMessage(GetDlgItem(hdlg, IDC_EF1), TBM_SETPOS, (WPARAM)TRUE, -ef1old);
					SendMessage(GetDlgItem(hdlg, IDC_EF2), TBM_SETPOS, (WPARAM)TRUE, -ef2old);
					SendMessage(GetDlgItem(hdlg, IDC_EF3), TBM_SETPOS, (WPARAM)TRUE, -ef3old);
				}
				if (redo) {
					SendMessage(GetDlgItem(hdlg, IDC_EF1), TBM_SETPOS, (WPARAM)TRUE, -ef1);
					SendMessage(GetDlgItem(hdlg, IDC_EF2), TBM_SETPOS, (WPARAM)TRUE, -ef2);
					SendMessage(GetDlgItem(hdlg, IDC_EF3), TBM_SETPOS, (WPARAM)TRUE, -ef3);
					SetDlgItemInt(hdlg, IDC_EF1VAL, ef1, TRUE);
					SetDlgItemInt(hdlg, IDC_EF2VAL, ef2, TRUE);
					SetDlgItemInt(hdlg, IDC_EF3VAL, ef3, TRUE);
					setFlg(mfd);
					mfd->ifp->RedoFrame();
				}
				undo = false;
				redo = false;
				break;
			}
	}

	return FALSE;
}

int ConfigProc(FilterActivation *fa, const FilterFunctions *ff, HWND hwnd)
{
	MyFilterData *mfd = (MyFilterData *)fa->filter_data;
	MyFilterData mfd_old = *mfd;
	int ret;

	mfd->ifp = fa->ifp;

	if (DialogBoxParam(fa->filter->module->hInstModule, MAKEINTRESOURCE(IDD_FILTER),
		hwnd, ConfigDlgProc, (LPARAM)mfd))
	{
		*mfd = mfd_old;
		ret = TRUE;
	}
    else
	{
		ret = FALSE;
	}
	return(ret);
}

void StringProc(const FilterActivation *fa, const FilterFunctions *ff, char *str)
{
	MyFilterData *mfd = (MyFilterData *)fa->filter_data;
}
 

void setFlg ( MyFilterData *mfd ) {
	mfd->shiFlg =(
			(mfd->rshpt != 0) || (mfd->gshpt != 0) || (mfd->bshpt != 0)
		||	(mfd->rshxx != 0) || (mfd->gshxx != 0) || (mfd->bshxx != 0)
		||	(mfd->rshyy != 0) || (mfd->gshyy != 0) || (mfd->bshyy != 0));
		switch (mfd->nmfunc) {
		case IDC_RSH:
		case IDC_GSH:
		case IDC_BSH:
			mfd->rshxy = mfd->rshyy * (mfd->width * 4 + mfd->salign) + mfd->rshxx * 4 + mfd->rshpt * 4;
			mfd->gshxy = mfd->gshyy * (mfd->width * 4 + mfd->salign) + mfd->gshxx * 4 + mfd->gshpt * 4;
			mfd->bshxy = mfd->bshyy * (mfd->width * 4 + mfd->salign) + mfd->bshxx * 4 + mfd->bshpt * 4;
			break;

	}
}


int normpos( int a, int b, int c, int dpos, long width) {
int ret;
if (dpos >0) {
		ret = a;
		if (b > ret) ret = b;
		if (c > ret) ret = c;
		if ((ret + dpos) < (width + 1)) ret = dpos;
		else ret = width - ret;
}
else {
		ret = a;
		if (b < ret) ret = b;
		if (c < ret) ret = c;
		if ((ret + dpos) > -(width + 1)) ret = dpos;
		else ret = -width - ret;
}
return(ret);
}

inline void normc(long *c) {
	if (*c < 0) *c = 0;
	if (*c > 255) *c = 255;
}

inline void normrgb(long *r, long *g, long *b) {
	if (*r < 0) *r = 0;
	if (*r > 255) *r = 255;
	if (*g < 0) *g = 0;
	if (*g > 255) *g = 255;
	if (*b < 0) *b = 0;
	if (*b > 255) *b = 255;
}


inline void shiftch(MyFilterData *mfd, long *r, long *g, long *b, Pixel32 *src,  int x, int y) {
	long udp, ri, gi, bi, uri, ugi, ubi;
	UCHAR *srcl;
	udp = (*r + *g + *b)/3;
	bool nmiddle, pre, post;

	pre = (x < (mfd->rshxx + mfd->rshpt));
	post = (x >= (mfd->width + mfd->rshxx + mfd->rshpt));
	nmiddle = (y < mfd->rshyy)
		|| (y >= (mfd->height + mfd->rshyy))
		|| (x < mfd->rshxx)
		|| (x >= (mfd->width + mfd->rshxx))
		|| (pre && (y <= mfd->rshyy))
		|| (post && (y >= (mfd->height - 1 + mfd->rshyy)));
	if (nmiddle) {
		ri = udp;
		uri = ri;
	}
	else {
		srcl = (UCHAR *)src - mfd->rshxy;
		if (pre) srcl -= mfd->salign;
		if (post) srcl += mfd->salign;
		ri = srcl[2];
		uri = (ri + srcl[1] + srcl[0])/3;
	}

	pre = (x < (mfd->gshxx + mfd->gshpt));
	post = (x >= (mfd->width + mfd->gshxx + mfd->gshpt));
	nmiddle = (y < mfd->gshyy)
		|| (y >= (mfd->height + mfd->gshyy))
		|| (x < mfd->gshxx)
		|| (x >= (mfd->width + mfd->gshxx))
		|| (pre && (y <= mfd->gshyy))
		|| (post && (y >= (mfd->height - 1 + mfd->gshyy)));
	if (nmiddle) {
		gi = udp;
		ugi = gi;
	}
	else {
		srcl = (UCHAR *)src - mfd->gshxy;
		if (pre) srcl -=mfd->salign;
		if (post) srcl +=mfd->salign;
		gi = srcl[1];
		ugi = (srcl[2] + gi + srcl[0])/3;
	}

	pre = (x < (mfd->bshxx + mfd->bshpt));
	post = (x >= (mfd->width + mfd->bshxx + mfd->bshpt));
	nmiddle = (y < mfd->bshyy)
		|| (y >= (mfd->height + mfd->bshyy))
		|| (x < mfd->bshxx)
		|| (x >= (mfd->width + mfd->bshxx))
		|| (pre && (y <= mfd->bshyy))
		|| (post && (y >= (mfd->height - 1 + mfd->bshyy)));
	if (nmiddle) {
		bi = udp;
		ubi = bi;
	}
	else {
		srcl = (UCHAR *)src - mfd->bshxy;
		if (pre) srcl -= mfd->salign;
		if (post) srcl += mfd->salign;
		bi = srcl[0];
		ubi = (srcl[2] + srcl[1] + bi)/3;
	}

	if (mfd->keeppt) {
		*r = ri - uri + udp;
		*g = gi - ugi + udp;
		*b = bi - ubi + udp;
		}
	else {
		*r = ri;
		*g = gi;
		*b = bi;
	}
	normrgb(r, g, b);
}

inline int normv(int v, int norm) {
if ( v > norm) return norm;
else return v;
if ( v < -norm) return -norm;
else return v;
}

UCHAR valtoch(int v, int ch) {
	UCHAR ret = 0;
	switch (ch) {
	case 1:
		if (v < 0) ret = 110;
		else ret = 112;
		return ret;
	case 2:
		ret = (UCHAR)(abs(v) + 128);
		return ret;
	}
	return ret;
}

int chtoval(UCHAR ch1, UCHAR ch2) {
	int ret = 0;
	switch (ch1) {
	case 110:
		ret = -(int)(ch2 - 128);
		return ret;
	case 112:
		ret = (int)(ch2 - 128);
		return ret;
	}
	return ret;
}
